/*

MIT License

Copyright (c) 2018 Kenichi Koizumi

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include <cinttypes>
#include "cvp.h"
#include <cassert>
#include <cstdio>
#include <stdlib.h>
#include <deque>
#include <functional>
#include "mypredictor.h"

HistoryTable htable;
MyPredictor pred1;
MyPredictor pred2;
MyPredictor pred3;

bool getPrediction(uint64_t seq_no, uint64_t pc, uint8_t piece,
                   uint64_t &predicted_value) {
    if (predictor_predict(&htable, &pred1, pc)) {
        predicted_value = predictor_get(&htable, &pred1, pc);
        return true;
    }
    if (predictor_predict(&htable, &pred2, pc)) {
        predicted_value = predictor_get(&htable, &pred2, pc);
        return true;
    }
    if (predictor_predict(&htable, &pred3, pc)) {
        predicted_value = predictor_get(&htable, &pred3, pc);
        return true;
    }
    return false;
}

void speculativeUpdate(uint64_t seq_no, bool eligible,
                       uint8_t prediction_result, uint64_t pc, uint64_t next_pc,
                       InstClass insn, uint8_t piece, uint64_t src1,
                       uint64_t src2, uint64_t src3, uint64_t dst) {
    uint64_t idx = hash(pc, param_entry_width);
    if (eligible) {
        htable.inflightPreds.push_front({seq_no, pc});
    }
    if (prediction_result == 0) {
        if (predictor_predict(&htable, &pred1, pc)) {
            predictor_init(&pred1, pc);
            pred1.failed[idx] = 1;
            return;
        }
        if (predictor_predict(&htable, &pred2, pc)) {
            predictor_init(&pred2, pc);
            pred2.failed[idx] = 1;
            return;
        }
        if (predictor_predict(&htable, &pred3, pc)) {
            predictor_init(&pred3, pc);
            pred3.failed[idx] = 1;
            return;
        }
    }
}

void updatePredictor(uint64_t seq_no, uint64_t actual_addr,
                     uint64_t actual_value, uint64_t actual_latency) {
    if (htable.inflightPreds.size() &&
        seq_no == htable.inflightPreds.back().seqNum) {
        uint64_t pc = htable.inflightPreds.back().pc;
        uint64_t idx = hash(pc, param_entry_width);
        if (preUpdateHistoryTable(&htable, pc)) {
            predictor_init(&pred1, pc);
            predictor_init(&pred2, pc);
            predictor_init(&pred3, pc);
            pred1.failed[idx] = 0;
            pred2.failed[idx] = 0;
            pred3.failed[idx] = 0;
        }
        predictor_update(&htable, &pred1, seq_no, actual_value, pc);
        predictor_update(&htable, &pred2, seq_no, actual_value, pc);
        predictor_update(&htable, &pred3, seq_no, actual_value, pc);
        postUpdateHistoryTable(&htable, pc, actual_value);
        htable.inflightPreds.pop_back();
    }
}

void beginPredictor(int argc_other, char **argv_other) {
    param_entry_width = 7;
    param_conf_threshold_1st = 24;
    param_conf_threshold_2nd = 112;

    initializeHistoryTable(&htable);
    predictor_init(&pred1, TYPE_DIFF, 1);
    predictor_init(&pred2, TYPE_PRDC, 2);
    predictor_init(&pred3, TYPE_PRDC, 3);
    printf("  [Info] beginPredictor(): done.\n");
}

void endPredictor() {
    freeHistoryTable(&htable);
    predictor_free(&pred1);
    predictor_free(&pred2);
    predictor_free(&pred3);
}
