/*

MIT License

Copyright (c) 2018 Kenichi Koizumi

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#define STATE_INIT 0
#define STATE_PREPARE1 1
#define STATE_PREPARE2 2
#define STATE_VERIFY 3
#define STATE_SPECULATE 4
#define TYPE_PRDC 0
#define TYPE_DIFF 1

int param_conf_threshold_1st;
int param_conf_threshold_2nd;
int param_entry_width;

struct HistoryTable {
    uint64_t *pc = NULL;
    uint64_t **history = NULL;
    int64_t *diff = NULL;
    uint64_t *newpc = NULL;
    char *newpc_cnt = NULL;
    struct InflightInfo {
        uint64_t seqNum;
        uint64_t pc;
        InflightInfo() : seqNum(0), pc(0) {}
        InflightInfo(uint64_t sn, uint64_t pcn) : seqNum(sn), pc(pcn) {}
    };
    std::deque<InflightInfo> inflightPreds;
};

struct MyPredictor {
    int histlen;
    int type;
    int periodlen;
    char *status = NULL;
    unsigned char *count = NULL;
    char *failed = NULL;
};

uint64_t hash(uint64_t s, int w) {
    char str[256];
    sprintf(str, "%lu", s);
    return ((1UL << w) - 1) & (uint64_t)std::hash<std::string> {}
    (str);
}

void initializeHistoryTable(HistoryTable *ht) {
    int k;
    uint64_t lli;
    uint64_t entryNum = 1ULL << param_entry_width;
    ht->pc = (uint64_t *)malloc(sizeof(uint64_t) * entryNum);
    ht->history = (uint64_t **)malloc(sizeof(uint64_t *) * 3);
    for (k = 0; k < 3; k++) {
        ht->history[k] = (uint64_t *)malloc(sizeof(uint64_t) * entryNum);
    }
    ht->diff = (int64_t *)malloc(sizeof(uint64_t) * entryNum);
    ht->newpc = (uint64_t *)malloc(sizeof(uint64_t) * entryNum);
    ht->newpc_cnt = (char *)malloc(sizeof(char) * entryNum);

    if ((ht->pc == NULL) || (ht->history == NULL) || (ht->diff == NULL) ||
        (ht->newpc == NULL) || (ht->newpc_cnt == NULL)) {
        printf("  [Error] malloc() failed.\n");
        exit(EXIT_FAILURE);
    }
    for (k = 0; k < 3; k++) {
        if (ht->history[k] == NULL) {
            printf("  [Error] malloc() failed.\n");
            exit(EXIT_FAILURE);
        }
    }
    for (lli = 0; lli < entryNum; lli++) {
        ht->pc[lli] = 0xdeadbeef;
        for (k = 0; k < 3; k++) {
            ht->history[k][lli] = 0xdeadbeef;
        }
        ht->diff[lli] = 0;
        ht->newpc[lli] = 0xdeadbeef;
        ht->newpc_cnt[lli] = 0;
    }
}

void predictor_init(MyPredictor *p, int type, int periodlen) {
    uint64_t lli;
    p->histlen = param_conf_threshold_2nd;
    p->type = type;
    p->periodlen = periodlen;
    uint64_t entrylen = 1ULL << param_entry_width;

    if ((periodlen < 1) || (3 < periodlen)) {
        printf("  [Error] period is out of range.\n");
        exit(EXIT_FAILURE);
    }
    if ((type == TYPE_DIFF) && (periodlen != 1)) {
        printf("  [Error] bad period for TYPE_DIFF.\n");
        exit(EXIT_FAILURE);
    }

    p->status = (char *)malloc(sizeof(char) * entrylen);
    p->count = (unsigned char *)malloc(sizeof(unsigned char) * entrylen);
    p->failed = (char *)malloc(sizeof(char) * entrylen);

    if ((p->status == NULL) || (p->count == NULL) || (p->failed == NULL)) {
        printf("  [Error] malloc() failed.\n");
        exit(EXIT_FAILURE);
    }

    for (lli = 0; lli < entrylen; lli++) {
        p->status[lli] = STATE_INIT;
        p->failed[lli] = 0;
    }
}

void freeHistoryTable(HistoryTable *ht) {
    int k;
    free(ht->pc);
    for (k = 0; k < 3; k++) {
        free(ht->history[k]);
    }
    free(ht->history);
    free(ht->diff);
    free(ht->newpc);
    free(ht->newpc_cnt);
}

void predictor_free(MyPredictor *p) {
    free(p->status);
    free(p->count);
    free(p->failed);
}

int predictor_queued(const HistoryTable *ht, uint64_t pc) {
    int cnt = 0;
    for (auto itr = ht->inflightPreds.begin(); itr != ht->inflightPreds.end();
         itr++) {
        if (pc == (*itr).pc) {
            cnt++;
        }
    }
    return cnt;
}

void predictor_init(MyPredictor *p, uint64_t pc) {
    uint64_t idx = hash(pc, param_entry_width);
    p->status[idx] = STATE_INIT;
}

bool preUpdateHistoryTable(HistoryTable *ht, uint64_t pc) {
    uint64_t idx = hash(pc, param_entry_width);
    uint64_t e_pc = ht->pc[idx];
    uint64_t e_newpc = ht->newpc[idx];
    char e_newpc_cnt = ht->newpc_cnt[idx];

    if (e_pc == 0xdeadbeef) {
        ht->pc[idx] = pc;
        ht->history[0][idx] = 0xdeadbeef;
        ht->history[1][idx] = 0xdeadbeef;
        ht->history[2][idx] = 0xdeadbeef;
        ht->diff[idx] = 0xdeadbeef;
        ht->newpc_cnt[idx] = 0;
        return true;
    }
    if (e_pc == pc) {
        return false;
    }
    if ((e_newpc == pc) && (e_newpc_cnt == 2)) {
        ht->pc[idx] = pc;
        ht->history[0][idx] = 0xdeadbeef;
        ht->history[1][idx] = 0xdeadbeef;
        ht->history[2][idx] = 0xdeadbeef;
        ht->diff[idx] = 0xdeadbeef;
        ht->newpc_cnt[idx] = 0;
        return true;
    }
    return false;
}

void postUpdateHistoryTable(HistoryTable *ht, uint64_t pc,
                            uint64_t actual_value) {
    uint64_t idx = hash(pc, param_entry_width);
    uint64_t e_pc = ht->pc[idx];
    uint64_t e_newpc = ht->newpc[idx];
    char e_newpc_cnt = ht->newpc_cnt[idx];

    if (e_pc == pc) {
        ht->diff[idx] = actual_value - ht->history[0][idx];
        ht->history[2][idx] = ht->history[1][idx];
        ht->history[1][idx] = ht->history[0][idx];
        ht->history[0][idx] = actual_value;
        ht->newpc_cnt[idx] = 0;
        return;
    }
    if (e_newpc_cnt == 0) {
        ht->newpc[idx] = pc;
    } else if (e_newpc != pc) {
        ht->newpc[idx] = pc;
        ht->newpc_cnt[idx] = 0;
    }
    ht->newpc_cnt[idx]++;
}

void predictor_update(const HistoryTable *ht, MyPredictor *p, uint64_t seq_no,
                      uint64_t actual_value, uint64_t pc) {
    int k;
    uint64_t idx = hash(pc, param_entry_width);
    char e_status = p->status[idx];
    int e_periodlen = p->periodlen;

    uint64_t e_pc = ht->pc[idx];
    uint64_t e_history[e_periodlen];
    for (k = 0; k < e_periodlen; k++) {
        e_history[k] = ht->history[k][idx];
    }
    uint64_t last_value = (p->type == TYPE_DIFF) ? e_history[0] + ht->diff[idx]
                                                 : e_history[e_periodlen - 1];

    if (e_pc != pc) return;

    switch (e_status) {
        case STATE_INIT:
            p->status[idx] = STATE_PREPARE1;
            p->count[idx] = 1;
            break;
        case STATE_PREPARE1:
            p->status[idx] = STATE_PREPARE2;
            p->count[idx]++;
            break;
        case STATE_PREPARE2:
            p->status[idx] = STATE_VERIFY;
            p->count[idx]++;
            break;
        case STATE_VERIFY:
            if (actual_value == last_value) {
                p->count[idx]++;
                if (((p->failed[idx] == 0) &&
                     (p->count[idx] > param_conf_threshold_1st)) ||
                    ((p->failed[idx] == 1) && (p->count[idx] > p->histlen))) {
                    p->status[idx] = STATE_SPECULATE;
                }
            } else {
                p->status[idx] = STATE_INIT;
                p->failed[idx] = 1;
            }
            break;
        case STATE_SPECULATE:
            if (actual_value != last_value) {
                p->status[idx] = STATE_INIT;
                p->failed[idx] = 1;
            }
            break;
    }
}

bool predictor_predict(const HistoryTable *ht, const MyPredictor *p,
                       uint64_t pc) {
    uint64_t idx = hash(pc, param_entry_width);
    return ((p->status[idx] == STATE_SPECULATE) && (ht->pc[idx] == pc));
}

uint64_t predictor_get(const HistoryTable *ht, const MyPredictor *p,
                       uint64_t pc) {
    uint64_t idx = hash(pc, param_entry_width);
    int plen = p->periodlen;
    int queued = predictor_queued(ht, pc) + 1;
    int pidx = (plen - (queued % plen)) % plen;
    return ((p->type == TYPE_DIFF)
                ? ht->history[0][idx] + ht->diff[idx] * queued
                : ht->history[pidx][idx]);
}
